#include <vector>
#include <list>
#include "mex.h"
#include "wbt_detector.h"
#include "ThreadPool.h"

using namespace std;
using namespace WBTDetector;

void printDetectorInfo(const Detector & D)
{
    mexPrintf("dims: (%d,%d)\n", D.w, D.h);
    mexPrintf("nStages: %d\n", D.nStages);
    mexPrintf("nTrees: %d\n", D.nTrees);
    mexPrintf("nNodes: %d\n", D.nNodes);
    mexPrintf("splittp: %d\n", D.type);
}

Detector createDetectorFromStruct(const mxArray * f)
{
    Detector D;
    D.tp = (int*)mxGetData(mxGetField(f, 0, "tp"));
    D.splits = (float*)mxGetData(mxGetField(f, 0, "split"));
    D.children = (int*)mxGetData(mxGetField(f, 0, "children"));
    D.prob = (float*)mxGetData(mxGetField(f, 0, "prob"));
    int * dims = (int*)mxGetData(mxGetField(f, 0, "dims"));
    D.w = dims[0];
    D.h = dims[1];
    const int * d = mxGetDimensions(mxGetField(f, 0, "tp"));
    D.nNodes = d[0];
    D.nTrees = d[1];
    D.nStages = (int)mxGetScalar(mxGetField(f, 0, "n_stages"));
    // D.stgStart = (int*)mxGetData(mxGetField(f,0,"stg_start"));
    // D.stgEnd = (int*)mxGetData(mxGetField(f,0,"stg_end"));
    D.tau = (float*)mxGetData(mxGetField(f,0,"thr"));
    D.type = (int)mxGetScalar(mxGetField(f,0,"splittp"));
    return D;
}

#define TILE_SZ (32)

// lhs 0   1  2  3                rhs 0     1
//    [bbs,Ns,Nl,sgns] = MXSCANIMAGES(imgs, D)
void mexFunction(int nlhs, mxArray * lhs[], int nrhs, const mxArray * rhs[])
{
    const mxArray * mx_imgs = rhs[0];
    const mxArray * mx_detector = rhs[1];
    
    int nImgs = mxGetNumberOfElements(mx_imgs);
    
    Detector D = createDetectorFromStruct(mx_detector);
    // printDetectorInfo(D);
    
    ThreadPool pool(4);
    list< future<WBTDetector::Result> > results; // List of detection results for all threads
    vector<Image> ims(nImgs);
    
    for (int k = 0; k < nImgs; ++k)
    {
        const mxArray * img_k = mxGetCell(mx_imgs, k);
        if (!mxGetData(img_k)) break;
        ims[k] = Image((float*)mxGetData(img_k), mxGetNumberOfDimensions(img_k), mxGetDimensions(img_k));
        // mexPrintf("img: k=%d, w=%d, h=%d\n", k, ims[k].w, ims[k].h);
        const int max_x = ims[k].dims[0] - D.w - 1;
        const int max_y = ims[k].dims[1] - D.h - 1;
        if (max_x <= 0 || max_y <= 0) break;
        for (int y = 0; y < max_y; y+=TILE_SZ)
            for (int x = 0; x < max_x; x+=TILE_SZ)
            {
            //mexPrintf("x=%i:%i; y=%i:%i\n",x, min(x+TILE_SZ-1, max_x), y, min(y+TILE_SZ-1, max_y));
            results.emplace_back(
                    pool.enqueue(
                    scanImage, std::cref(ims[k]), x, y, min(x+TILE_SZ, max_x), min(y+TILE_SZ, max_y), std::cref(D), k+1
                    )
                    );
            }
    }
    
    vector<BB> bbs;
    vector<float> hs;
    vector<WBTDetector::Signature> ss;
    unsigned long long nf = 0;
    unsigned long long ns = 0;
    for (auto && result: results)
    {
        WBTDetector::Result r = result.get();
        bbs.insert(bbs.end(), r.bbs.begin(), r.bbs.end());
        hs.insert(hs.end(), r.hs.begin(), r.hs.end());
        ss.insert(ss.end(), r.ss.begin(), r.ss.end());
        ns += r.ns; nf += r.nf;
    }
    
    // Convert bb to 6xL array with bounding boxes (in intermediate format
    // [x,y,w,h,response,scale]).  This bonding boxes must be immediately
    // processed to handle image padding and scale
    lhs[0] = mxCreateNumericMatrix(6, bbs.size(), mxDOUBLE_CLASS, mxREAL);
    lhs[1] = mxCreateDoubleScalar(double(ns));
    lhs[2] = mxCreateDoubleScalar(double(nf));
    int sgnsz = min(D.nStages, WBTDetector::SignatureSize);
    lhs[3] = mxCreateNumericMatrix(sgnsz, bbs.size(), mxUINT8_CLASS, mxREAL);
    
    double * bbs_out = (double*)mxGetData(lhs[0]);
    for (size_t i = 0; i < bbs.size(); ++i)
    {
        BB & b = bbs[i];
        float h = hs[i];
        bbs_out[6*i+0] = b.y;
        bbs_out[6*i+1] = b.x;
        bbs_out[6*i+2] = b.height;
        bbs_out[6*i+3] = b.width;
        bbs_out[6*i+4] = h;
        bbs_out[6*i+5] = b.scale;
        
        unsigned char * sgnout=(unsigned char*)mxGetData(lhs[3]) + sgnsz * i; // i-th column of signature
        copy(ss[i].begin()+WBTDetector::SignatureSize-sgnsz, ss[i].end(), sgnout);
    }
}

// End