function i = weighted_sampling(method, N, W, varargin)
% Select N weighed samples
%
%   I = WEIGHTED_SAMPLING(METHOD, N, W, ...);
%
% Select N samples according to their weights W using METHOD.
%
% METHOD can be:
% 'uus'   Select N samples with uniform distribution (disregarding the
%         weights). This method corresponds to randsample function.
% 'trim'  Take N samples with highest weights.
% 'qws'   Quasi-random weighted sampling [1].
% 'qws+'  Quasi-random weighted sampling + trimmig [1].
%
% Additional arguments:
% unique  [false] Remove duplicate samples (for qws and qws+)
% L       [0] Number fo samples to trim (for qws+). 0<=L<=N.
%         For L=0 the method converge to qws, and L==N, it converge to
%         trim.
%
% Example:
%   ind = weighted_sampling('qws+', 50, W, 'L', 10, 'unique', true);
%
% Notes and Known Issues:
% * The methods are class-agnostic. They do not take class labels into
% account in any way.
% * When the weights are equal, trim returns first N items. This may cause
% problems when samples are ordered in some way. For example, when first
% half of dataset consists only form positive samples, and the second half
% from negative. Trimming N<length(W)/2 samples will result in one class
% subset, while one would rather sample both classes evenly. Remedy to this
% is threefold. 1/ sample classes independently; or 2/ randomly reorder items
% prior to sampling; or 3/ use uus method instead.
% * qws+ actually calls trim and qws in sequence and so it is a bit slower
% than an optimized implementation would be.
%
% [1] Kalal,Z., et al.: Weighted Sampling for Large-Scale Boosting, BMVC 2008
%
% See also: RANDSAMPLE

opts = struct(varargin{:});
N = round(N);

switch method
    case {'uus'},
        i = randsample(length(W), N);
        
    case {'trim'},
        if N > length(W), i = 1:length(W);
        else
            [~,ord] = sort(W,'descend');
            i = ord(1:N);
        end
        
    case {'qws'},
        if ~isrow(W), W = W'; end;
        W = W./sum(W);
        % Interval limits
        lim1 = cumsum(W); lim0 = [0 lim1(1:end-1)];
        % Generate points in L even intervals
        j = (0:N-1)/N + rand(1,N)/N;
        % Get the samples - check to which interval falls j
        [i,~] = find(bsxfun(@ge,j,lim0') & bsxfun(@lt,j,lim1'));
        if isfield(opts,'unique')&&opts.unique~=0, i = unique(i); end;
        
    case {'qws+'},
        if length(W) < opts.L,
            i = 1:length(W);
        else
            i1 = weighted_sampling('trim', opts.L, W);
            K = N - opts.L;
            rest = setdiff(1:length(W), i1);
            i_rest = weighted_sampling('qws', K, W(rest), varargin{:});
            i2 = rest(i_rest)';
            i = cat(1, i1, i2);
        end
    otherwise,
        error('Unknown method');
end

