function [D,DATA] = wbTrainDetector(opts)
% Train object detector
%
%   OPTS = TRAIN_DETECTOR;
%   D = TRAIN_DETECTOR(OPTS);


if nargin == 0
    D.pTree = random_forest_train;
    D.pPyramid = chnsPyramid;
    D.pJitter = [];
    D.imreadf = @imread;
    D.imreadp = {};
    D.posImgDir = '';
    D.posGtDir = '';
    D.posVptDir = '';
    D.negImgDir = '';
    D.negGtDir = '';
    D.samples = []; % or struct('file', 'data.mat', 'vars', {'X','V'});
    D.pLoadNeg = {};
    D.pLoadPos = {};
    D.nPos = 5000;
    D.nNeg = 20000;
    D.nPerPos = 1;
    D.nPerNeg = 20;
    D.modelDs = [];  % [height, width]
    D.modelDsPad = [];  % [height, width]
    %D.alpha = 0.05; % Taget false negative rate
    %D.beta = 0; % Target false positive rate (NOT USED NOW)
    %D.target_neg_rejection = 1; % Target true negative rate
    D.show_stats = false;
    D.nWeak = [];
    D.save = struct('file','detector.mat','every', 1);
    D.logfile = [];
    %D.term_loss = 1e-6; % Stop after achieving loss
    D.pose_model = [];
    % other pose model params?
    return;
end % default values


t_start = tic;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Create empty detector or re-use existing one
%%% -> D, opts, stg

if isfield(opts, 'opts') % opts is actually pre-trained detector containing options.
    D = opts;
    opts = opts.opts;
    stg = D.stage_info;
else % opts is really options - so we train new detector
    D = struct( ...
        'classes',[1 2], ...
        'split_eval', {opts.pTree.split_eval(1:2:end)}, ...
        'split_params', {opts.pTree.split_eval(2:2:end)}, ...
        'n_predictors', 1, ...
        'tree',{{}}, ...
        'tau',[], ...
        'opts', opts, ...
        'stage_info', [] ...
        );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Setup log file
%%% -> log

log = 0;
if isfield(opts, 'logfile') && ~isempty(opts.logfile)
    logprintf('Using %s as log file', opts.logfile);
    diary(opts.logfile);
    log = 1;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Setup basic training info
%%% -> sampling, train_pose_model, pose_model_start

sampling = true;
train_pose_model = isstruct(opts.pose_model);
pose_model_start = inf;
if train_pose_model
    pose_model_start = opts.pose_model.start;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load positive data
%%% -> X, Y, H, P

% Samples:
% The training keeps a pool of samples defined by variables:
% X     MxNxCxK matrix with image data (MxNxC for every sample)
% Y     Kx1 vector of labels
% P     KxL or [] matrix of sample poses
% H     Kx1 vector of responses (calculated automatically by the detector)
% W     Kx1 vector of sample weigths (calculated automatically from H)

[X,Y,V1,H,P] = deal([]);

% Load positive samples
if ~isempty(opts.posImgDir) && ~isempty(opts.posGtDir)
    logprintf('Loading positive data');
    
    if ~isempty(opts.posVptDir)
        %pos_opts.fs = bbGt('getFiles', {opts.posImgDir, opts.posGtDir, opts.posVptDir});
        pos_opts.fs = enum_files(opts.posImgDir, opts.posGtDir, opts.posVptDir);
    else
        %pos_opts.fs = bbGt('getFiles', {opts.posImgDir, opts.posGtDir});
        pos_opts.fs = enum_files(opts.posImgDir, opts.posGtDir);
    end
    
    pos_opts.imreadp = opts.imreadp;
    pos_opts.imreadf = opts.imreadf;
    pos_opts.pPyramid = opts.pPyramid;
    pos_opts.pJitter = opts.pJitter;
    pos_opts.pLoad = opts.pLoadPos;
    pos_opts.modelDs = opts.modelDs;
    pos_opts.modelDsPad = opts.modelDsPad;
    pos_opts.nPerPos = opts.nPerPos;
    pos_opts.random = true;
    
    diary off;
    [X,V1] = sample_from_gt(opts.nPos,pos_opts);
    diary on;
    if ~isempty(V1), V1 = V1(:,1:3); end;
    Y = 2*ones(size(X,4),1,'int32');
    H = zeros(size(X,4),1);
end

if ~isempty(opts.samples)
    if ischar(opts.samples)
        logprintf('Loading samples from %s', opts.samples);
        samples = load(opts.samples);
        loaded = true;
    elseif isstruct(opts.samples)
        logprintf('Loading samples from %s, variable %s', opts.samples.file, opts.samples.var);
        samples = load(opts.samples.file, opts.samples.var);
        samples = samples.(opts.samples.var);
        loaded = true;
    end

    if loaded % All samples are considered to be positive!
        logprintf('Loaded %i samples', size(samples.X,4));
        V1 = cat(1, V1, samples.V);
        X = cat(4, X, samples.X);
        H = cat(1, H, zeros(size(samples.X,4),1));
        if isfield(samples,'Y')
            Y = cat(1, Y, samples.Y);
        else
            Y = cat(1, Y, 2*ones(size(samples.X,4),1,'int32'));
        end
    else
        warning('No samples loaded');
    end
end

if ~isempty(V1) && train_pose_model
    logprintf('Calculating pseudo-residuals');
    if opts.pose_model.classes
        P = V1;
        if isfield(opts.pose_model, 'fProcess')
            P = feval(opts.pose_model.fProcess, P);
        end
    else
        P = zeros(size(X,4), size(opts.pose_model.V,1), 'single');
        if ~isfield(opts.pose_model, 'alpha'), opts.pose_model.alpha = 0.2; end;
        V1 = normr(V1);
        pm_V = opts.pose_model.V;
        for i = 1:size(X,4)
            [~,j] = max(V1(i,:) * pm_V');
            m = rasterVectors(pm_V(j,:), pm_V, 1, opts.pose_model.alpha);
            P(i,:) = m(:)';
        end % init for each sample
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% If detector exists and pose model is already partially trained, compute
%%% pseudo residuals - update P
%%% -> P

if ~isempty(P) && train_pose_model && numel(D.tree) >= pose_model_start
    logprintf('Updating pseudo-residuals');
    [~,~,I] = eval_detector(D, X);
    P = bsxfun(@minus, P, D.pose_model.init);
    for t = pose_model_start:numel(D.tree)
        pose_stage = opts.nWeak - pose_model_start + 1;
        for n = find([D.tree{t}.tp]==0)
            P(I{t}{n},:) = bsxfun(@minus, P(I{t}{n},:), D.pose_model.map{pose_stage}(n,:));
        end % nodes
    end  % trees
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Existing detector - filter data
%%% -> X, Y, H, P

if ~isempty(D.tree)
    [H,ind,~] = eval_detector(D, X);
    X = X(:,:,:,ind);
    Y = Y(ind);
    H = H(ind);
    if ~isempty(P), P = P(ind,:); end;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% If too much positive data was loaded - reject some
%%% -> X, Y, H, P

pos_ind = find(Y==2);
if length(pos_ind) > opts.nPos
    ind = sort(randsample(length(pos_ind), length(pos_ind)-opts.nPos));
    X(:,:,:,ind) = [];
    Y(ind) = [];
    H(ind) = [];
    if ~isempty(P), P(ind,:) = []; end;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Open window for training statistics
%%% -> f

if opts.show_stats
    f = figure;
    f.ToolBar = 'none';
    f.MenuBar = 'none';
    f.Name = 'Stage properties';
    f.Color = [1 1 1];
    f.Position = [1000 500 640 480];
    drawnow;
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Prepare stuff for background sampling
%%% -> neg_opts, neg_sample_info

if ~isempty(opts.negGtDir)
    neg_opts.fs = enum_files(opts.negImgDir, opts.negGtDir);
else
    neg_opts.fs = enum_files(opts.negImgDir);
end

neg_opts.imreadp = opts.imreadp;
neg_opts.imreadf = opts.imreadf;
neg_opts.pPyramid = opts.pPyramid;
neg_opts.dim = opts.modelDsPad / opts.pPyramid.pChns.shrink;
neg_opts.nPerImage = opts.nPerNeg;
neg_opts.pLoad = opts.pLoadNeg;
neg_opts.detector = [];
neg_sample_info(1:size(neg_opts.fs,2)) = struct('loaded',0,'R',[],'left',nan);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% MAIN TRAINING LOOP STARTS HERE - iteratively update D, X, Y, P,...

for stage = (numel(D.tree)+1):opts.nWeak;
    stage_t0 = tic; % Mark start
    stg(stage).id = stage; %#ok<*AGROW>
    
    logprintf(['Stage %d/%d ' repmat('-',1,40)], stage, opts.nWeak);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% If needed, load more background samples - bootstrapping
    %%% -> X, Y, H, P, sampling
    if sampling
        if stage > 1, neg_opts.detector = preprocess_detector(D); end;
        need_samples = opts.nNeg - sum(Y==1);
        if need_samples > 0
            logprintf('Loading negative data');
            diary off;
            [newX0,newH0,neg_sample_info] = sample_windows(need_samples, neg_opts, neg_sample_info);
            diary on;
            
            newY0 = ones(size(newH0),'int32');
            newP0 = [];
            
            % Integrity checking
            [newH0_check,newH0_ind] = eval_detector(D, newX0);
            if length(newH0_ind) < length(newH0_check) || any(newH0 ~= newH0_check)
                % This should never happen - samples loaded by image
                % scanning MUST pass detector evaluation which does
                % exactly the same computation. If some samples do not pass
                % or has different response, it means that MEX version of
                % detector makes mistakes for some reason. eval_detector is
                % ALWAYS RIGHT as it is used for training.
                
                % THIS IS JUST A HACK! REMOVE SAMPLES THAT DIFFER IN
                % RESPONSE. THIS DEFINITELY SHOULD NOT HAPPEN!
                rem_h = newH0 ~= newH0_check;
                rem_i = true(size(newH0));
                rem_i(newH0_ind) = false;
                rem = rem_h | rem_i;
                newX0(:,:,:,rem) = [];
                newH0(rem) = [];
                newY0(rem) = [];
                if ~isempty(newP0), newP0(rem) = []; end;
                
                logprintf('Integrity check - %i samples discarded', sum(rem));
                %error('Integrity error');
            end
            
            if ~isempty(P)
                newP0 = zeros(size(newH0,1), size(P,2), 'single');
            end
            
            % Add new samples to dataset
            if ~isempty(newX0)
                X = cat(4, X, newX0);
                Y = cat(1, Y, newY0);
                H = cat(1, H, newH0);
                P = cat(1, P, newP0);
            end
            % Clear tmps
            clear newX0 newY0 newH0 newV0 newP0;
        end
        if stage > 1, sampling = (sum(Y==1) >= opts.nNeg); end;
    end % sampling
    
    % This is the actual number of samples loaded from background dataset.
    % Number of samples that were analyzed (during dataset scanning) is far
    % larger than this.
    logprintf('Total # loaded negative samples: %i', sum([neg_sample_info.loaded]));
    %logprintf('Estimated samples left per image: %.1f', nansum([neg_sample_info.left]) / sum(~isnan([neg_sample_info.left])));
    %logprintf('# exhausted images: %i', nansum([neg_sample_info.left] == 0));
    
    if ~sampling, logprintf('Sampling stopped'); end;
    stg(stage).sampling = sampling;
    
    logprintf('Pool: %d samples (#neg: %d, #pos: %d)', length(Y), histc(Y,[1,2]));
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Calc weight
    %%% -> W
    W = weight(H, Y);
    
    stg(stage).J = mean(W); % Current loss
    logprintf('Loss: %e', stg(stage).J);
    
    if any(~isfinite(W)) || any(~isfinite(H)), error('H and W must be finite!'); end;
    
    neg_ind = Y == 1;
    pos_ind = Y == 2;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Select subset of training data for stage training
    %%% -> X1, Y1, W1, P1
    if stage==1, sampling_type = 'uus';
    else sampling_type = 'qws+'; end;
    ind = subsample_dataset(opts.subsample, Y, W, P, sampling_type);
    
    X1 = X(:,:,:,ind);
    Y1 = Y(ind);
    W1 = W(ind);
    if ~isempty(P),
        P1 = P(ind,:);
    end % P1 must be used in error minimization in random_forest_train
    
    %{
    [~,w_ord] = sort(W1, 'ascend');
    w_ord = w_ord(1:length(w_ord)/2);
    w_ord = randsample(w_ord, round(length(w_ord)/10));
    W1(w_ord) = 10 * W1(w_ord);
    %}
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Stage training
    %%% -> tmp_forest, h, H, I, D
    
    logprintf('Training Random Forest');
    
    % Setup training options
    pTree = opts.pTree;
    pTree.classes = D.classes;
    pTree.ntrees = 1;
    pTree.subsample = 0;
    pTree.predictor = {'wb_predictor',{1e-5}};
    
    % Setup pTree for regression stages
    pTree.impurity = {'adaboost_error',{}};
    if train_pose_model && stage >= pose_model_start
        if isfield(opts.pose_model,'depth')
            pTree.max_depth = opts.pose_model.depth;
        end
        switch opts.pose_model.error
        case {'classification'}
            pTree.impurity = {'adaboost_error',{}};
        case {'regression'}
            pTree.impurity = {'pose_residual_error',{P1}};
        case {'combined'}
            pTree.impurity = {'adaboost_error',{},'pose_residual_error',{P1}};
        end
    end
    
    % Train the forest and update it using full set
    tmp_forest = random_forest_train(X1,Y1,W1,pTree);
    tmp_forest = random_forest_update(tmp_forest,X,Y,W,pTree.predictor{:});
    
    % Update sample responses
    logprintf('Updating sample responses');
    [h,I] = random_forest_predict(tmp_forest, X);
    H = H + h;
    if any(~isfinite(h)), error('Predicted values MUST be finite!'); end;
    
    % Add the trained trees (tmp_forest) to the detector D
    D.tree = horzcat(D.tree, tmp_forest.tree);
    
    % Set rejection threshold
    D.tau(stage) = -inf;
    if sampling && length(opts.thr) >= stage
        D.tau(stage) = opts.thr(stage);
    end
    logprintf('thr = %3.2f', D.tau(stage));
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Fit gradient boosting to P and update P with current prediction    
    % -> D.pose_model
    
    if train_pose_model && stage >= pose_model_start
        pose_stage = stage - pose_model_start + 1;
        
        logprintf('Pose model stage: %i', pose_stage);
        
        if pose_stage == 1
            logprintf('New pose model: size=%ix%i', size(opts.pose_model.V));
            D.pose_model.init = mean(P(pos_ind,:),1);
            D.pose_model.map = {};
            D.pose_model.V = opts.pose_model.V;
            P = bsxfun(@minus, P, D.pose_model.init);
        end
        
        I = I{1};
        M = zeros(numel(tmp_forest.tree{1}), size(P,2), 'single');
        for n = find([tmp_forest.tree{1}.tp] == 0)
            k = I{n}(Y(I{n}) == 2);
            m = 0.5 * bsxfun(@rdivide, sum(P(k,:),1), length(k)+1);
            M(n,:) = m;
            P(k,:) = bsxfun(@minus, P(k,:), m);
        end
        D.pose_model.map{pose_stage} = M;
        pr_err = P(pos_ind,:).^2;
        logprintf('Pseudo-residual error: %.2f', sum(pr_err(:)));
        %{
        figure(16);
        qq = find(pos_ind);
        subplot(3,1,1:2);
        montage2(reshape(P(qq(1:10:end),:)', 33,33,[]));
        colormap jet;
        colorbar;
        subplot(3,1,3);
        h_bins = linspace(0, max(pr_err(:)), 100);
        h_pr_err = cumsum(histc(pr_err(:), h_bins));
        plot(h_bins, h_pr_err/sum(h_pr_err(:)), '-');
        drawnow;
        hold off;
        %pause(0.5);
        %}
    end % pose model branch
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Get stage statistics
    %%% -> stg, D.stg
    
    H_bins = linspace(min(H)-1, max(H)+1, 64);
    h_bins = linspace(-6, 6, 64);
    
    stg(stage).H_bins = H_bins;
    stg(stage).neg_pdf = estimate_distribution(H(neg_ind), H_bins);
    stg(stage).pos_pdf = estimate_distribution(H(pos_ind), H_bins);
    
    stg(stage).h_bins = h_bins;
    stg(stage).neg_pdf_h = estimate_distribution(h(neg_ind), h_bins);
    stg(stage).pos_pdf_h = estimate_distribution(h(pos_ind), h_bins);

    stg(stage).tau = D.tau(stage);
    D.stage_info = stg;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Filter data
    %%% -> X. Y, H, W, P
    %%% W update is probably not necessary as it is recalulated at the
    %%% begining of every stage
    reject = H < D.tau(stage);
    X(:,:,:,reject) = [];
    Y(reject) = [];
    H(reject) = [];
    W(reject) = [];
    if ~isempty(P), P(reject,:) = []; end;
    
    % Print tool info
    logprintf('Pool: %d samples (#neg: %d, #pos: %d)', length(Y), histc(Y, [1,2]));
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Optionally show statistics
    if opts.show_stats
        figure(f);
        wbt.show_stage_stats(stg(stage))
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%% Save detector
    if mod(stage,opts.save.every)==0 || stage == opts.nWeak %|| quit
        logprintf('Saving %s', opts.save.file);
        save(opts.save.file, 'D');
    end
    
    stage_t1 = toc(stage_t0); logprintf('Stage time: %.1fs', stage_t1);   
    
end % stages

logprintf('Done (%is)', round(toc(t_start)));

if nargout==2
    DATA.X = X;
    DATA.Y = Y;
    DATA.H = H;
    DATA.W = W;
    DATA.P = P;
end

if log, diary off; end;

end % train_detector


function [pdf,cdf] = estimate_distribution(H,bins)
pdf = histc(H,bins);
pdf = pdf./sum(pdf);
cdf = cumsum(pdf);
end

function W = weight(H,Y)
W = zeros(size(H));
W(Y==1) = exp(H(Y==1));% ./ (sum(L==1)/2);
W(Y==2) = exp(-H(Y==2));% ./ (sum(L~=1)/2);
W = min(max(W,1e-50),1e50);
end

function ind = subsample_dataset(N, Y, W, ~, tp)
% Draw random samples from dataset

if N > length(Y)
    ind = 1:length(Y); % There is not enough samples - use all
else
    %%% TAKE BACKGROUND SAMPLES
    
    c_ind = find(Y == 1);
    i = weighted_sampling(tp, N/2, W(c_ind), 'L', N/4, 'unique', true);
    bg_ind = c_ind(i);
    
    %%% TAKE OBJECT SAMPLES
    c_ind = find(Y == 2);
    % 1. if there are vpts present, take uniform vpts
    %{
    if ~isempty(P)
        if size(P,2) == 3 % viewpoints
            P = normr(P);
            i = sample_viewpoints_nonuniform(P(c_ind,:), N);
            c_ind = c_ind(i);
        end
        if size(P,2) == 1 % subclasses
            % Stratified sampling - c_ind
            i = sample_stratified(P(c_ind,:), N);
            c_ind = c_ind(i);
        end
    end
    %}
    % Weighted sample
    i = weighted_sampling(tp, N/2, W(c_ind), 'L', N/4, 'unique', true);
    obj_ind = c_ind(i);
    
    ind = [bg_ind; obj_ind];
    
end
if ~issorted(ind), ind = sort(ind); end;
end

function logprintf(message, varargin)
%mfprintf = @(fid,msg,params)arrayfun(@(f)fprintf(f,msg,params{:}),fid,'uniformoutput',0);
message = ['[' datestr(now,31) ']: ' message '\n'];
fprintf(message, varargin{:});
end


function fs = enum_files(varargin)
% fs = get_all_files({A1,A2,A3},{B1,B2,B3});
dirsets = varargin;
for i = 1:numel(dirsets)
    if ~iscell(dirsets{i}), dirsets{i} = {dirsets{i}}; end;
end
fs = {};
for i = 1:numel(dirsets{1})
    X = cellfun(@(x)x{i},dirsets,'uniformoutput',0);
    fs{i} = bbGt('getFiles',X);
end
fs = horzcat(fs{:});
end

